# -*- coding: utf-8 -*-
import os
import sys
import math
import time
import datetime
import random

import c4d
from c4d import plugins
from c4d import documents
from c4d import gui
from c4d import utils

from smartstairspackage import stairobj
from smartstairspackage import worker

from smartstairspackage.utils import errormsg
from smartstairspackage.utils import smart
from smartstairspackage.utils import hivemind

from smartstairspackage import config
from smartstairspackage import ids


class SmartStairsData(plugins.ObjectData):
    
    HANDLECOUNT = 3
    
    platform_bb = c4d.Vector()
    platform_pos = c4d.Vector()

    linkedHack = False
    linkedHack2 = False
    linkedHack3 = False
    childHack = False
    parentHack = False
            
    selftest = False
    testtype = 0


    def __init__(self):

        hivemind.Borg().errorCallBack = self.errorCallBack
                        
        if hasattr(config, "SELFTEST"):
            self.selftest = True
            
        if hasattr(config, "TESTTYPE"):
            self.testtype = config.TESTTYPE #@UndefinedVariable
        
        
    def errorCallBack(self, a, aForce = False):
        
        if aForce or (not hivemind.Borg().fatalError and ("smartstairspackage" in a)):
            hivemind.Borg().fatalError = True


    def Message(self, aNode, aType, aData):
        
        if hivemind.Borg().fatalError:
            if not hivemind.Borg().fatalShown and config.SMARTSTAIRS_DEBUG:
                c4d.CallCommand(12305)
                hivemind.Borg().fatalShown = True
            return False
                
        if aType == c4d.MSG_DESCRIPTION_COMMAND:            
            if aData['id'][0].id == ids.SMARTSTAIRS_STAIRS_COPYBTN:
                hivemind.Borg().copy = aNode
                
            if aData['id'][0].id == ids.SMARTSTAIRS_STAIRS_PASTEBTN:
                if gui.QuestionDialog(plugins.GeLoadString(ids.STR_ASKPASTE)):
                    self.getParameters(aNode)
                    c4d.EventAdd()
                
            if aData['id'][0].id == ids.SMARTSTAIRS_STAIRS_CHILDBTN:
                newss = c4d.BaseObject(ids.PLUGIN_ID)
                doc = documents.GetActiveDocument()
                if doc is not None and doc.IsAlive():
                    doc.StartUndo()
                    doc.InsertObject(newss, aNode)
                    c4d.CallCommand(100004802)
                    self.childHack = False
                    doc.AddUndo(c4d.UNDOTYPE_NEW, newss)
                    doc.EndUndo()
                    
            if aData['id'][0].id == ids.SMARTSTAIRS_STAIRS_RESETBTN:
                if gui.QuestionDialog(plugins.GeLoadString(ids.STR_ASKRESETTODEFAULTS)):
                    stairobj.Stair(aNode).setDefaults()

        return True


    def Init(self, aNode):

        data = aNode.GetDataInstance()
        data.SetString(ids.SMARTSTAIRS_ABOUT_TXT, "Smart Stairs - version %s (2020 @ trideon.de / smart-page.net)" % ids.VERSION)
            
        stairobj.Stair(aNode).setDefaults()
            
        return True


    def GetHandle(self, aObj, aI, aInfo):
        
        if hivemind.Borg().fatalError: return

        stair = stairobj.Stair(aObj)

        val = 0
        vec = c4d.Vector(0)
        pos = c4d.Vector(0)

        if aI == 0:
            val = stair.width / 2.0
            vec = c4d.Vector(1.0, 0.0, 0.0)
            pos = c4d.Vector(val, 0.0, 0.0)
        if aI == 1:
            val = stair.height
            vec = c4d.Vector(0.0, 1.0, 0.0)
            pos = c4d.Vector(0.0, val, 0.0)
        if aI == 2:
            val = stair.depth / 2.0
            if stair.stair_anchor != ids.SMARTSTAIRS_STAIRS_ANCHOR_CENTER:
                val = stair.depth
                
            vec = c4d.Vector(0.0, 0.0, 1.0)
            
            if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                val = -stair.depth
                vec = c4d.Vector(0.0, 0.0, -1.0)

            pos = c4d.Vector(0.0, 0.0, val)

        aInfo.position = pos
        aInfo.direction = vec
        aInfo.type = c4d.HANDLECONSTRAINTTYPE_LINEAR


    def SetHandle(self, aObj, iI, aP, aInfo):
        
        if hivemind.Borg().fatalError: return

        stair = stairobj.Stair(aObj)
        
        val = aP * aInfo.direction

        if iI == 0:
            stair.width = utils.FCut(val * 2.0, 0.0, sys.maxint)

        if iI == 1:
            stair.height = utils.FCut(val, 0.0, sys.maxint)

        if iI == 2:
            if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_CENTER:
                val *= 2
                
            stair.depth = utils.FCut(val, 0.0, sys.maxint)


    def Draw(self, aObj, aDrawpass, aBd, aBh):
        
        if hivemind.Borg().fatalError: return c4d.DRAWRESULT_OK

        if aDrawpass != c4d.DRAWPASS_HANDLES: return c4d.DRAWRESULT_SKIP
        
        stair = stairobj.Stair(aObj)

        hitid = aObj.GetHighlightHandle(aBd)

        m = aBh.GetMg()
        aBd.SetMatrix_Matrix(aObj, m)

        hit = False
        
        info = c4d.HandleInfo()
    
        for i in range(SmartStairsData.HANDLECOUNT):
            if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX and i == 2:
                continue
                
            if hitid == i:
                col = c4d.GetViewColor(c4d.VIEWCOLOR_SELECTION_PREVIEW)
                hit = True
            else:
                col = c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT)

            aBd.SetPen(col)

            self.GetHandle(aObj, i, info)

            aBd.DrawHandle(info.position, c4d.DRAWHANDLE_BIG, 0)
            aBd.DrawLine(info.position, c4d.Vector(0.0), 0)

        if hit:
            col = c4d.GetViewColor(c4d.VIEWCOLOR_SELECTION_PREVIEW)
        else:
            col = c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT)

        offz = 0
        if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
            offz = stair.depth / 2.0
        if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
            offz = - stair.depth / 2.0
            
        width = stair.width
        height = stair.height
        depth = stair.depth

        m = c4d.Matrix()

        m.v1 *= width
        m.v2 *= height
        m.off = c4d.Vector(0.0, height / 2.0, offz)
        m.v3 *= depth
        
        aBd.DrawBox(m, 0.5, col, True)
        
        if stair.stair_tobottom and stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID:
            ground = 0
            mo = stair.obj.GetMg().off
            if stair.stair_bottomoffset < mo.y:
                ground = mo.y - stair.stair_bottomoffset
                
            if ground > 0:
                col = c4d.GetViewColor(c4d.VIEWCOLOR_ANIMPATH)
                
                m = c4d.Matrix()

                m.v1 *= width
                m.v2 *= ground
                m.off = c4d.Vector(0.0, - ground / 2.0, offz)
                m.v3 *= depth

                aBd.DrawBox(m, 0.5, col, True)
        
        if self.platform_bb.GetLength() > 0 and stair.is_parent:
            width = self.platform_bb.x
            height = self.platform_bb.y
            depth = self.platform_bb.z
                   
            m = c4d.Matrix()    
            m.v1 *= width
            m.v2 *= height
            m.v3 *= depth
            m.off = c4d.Vector(self.platform_pos.x, 
                               self.platform_pos.y - height / 2.0 + stair.treads_height, 
                               self.platform_pos.z + depth / 2.0 + offz - stair.depth / 2.0)            
            
            col = c4d.GetViewColor(c4d.VIEWCOLOR_ANIMPATH)
            aBd.DrawBox(m, 0.5, col, True)
            
        if stair.has_railing:
            width = stair.width
            height = stair.railing_height + stair.railing_rad2
            if stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
                hoff = stair.height / stair.depth * stair.railing_overtop
                if hoff > stair.railing_overtop * 4: hoff = stair.railing_overtop * 4
                height += hoff
                
            if stair.reverse and stair.is_child and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
                height -= stair.step_dims.y
                 
            depth = stair.depth
            
            m = c4d.Matrix()    
            m.v1 *= width
            m.v2 *= height
            m.v3 *= depth
            m.off = c4d.Vector(0.0, 
                               stair.height + height / 2.0, 
                               offz)            
            
            col = c4d.GetViewColor(c4d.VIEWCOLOR_ANIMPATH)
            aBd.DrawBox(m, 0.5, col, True)
        
        return c4d.DRAWRESULT_OK


    def DetectHandle(self, aObj, aBd, aX, aY, aQualifier):
        
        if hivemind.Borg().fatalError: return c4d.NOTOK

        if aQualifier & c4d.QUALIFIER_CTRL: return c4d.NOTOK

        mg = aObj.GetMg()
        
        ret = c4d.NOTOK

        for i in range(SmartStairsData.HANDLECOUNT):
            info = c4d.HandleInfo()
            self.GetHandle(aObj, i, info)
            if aBd.PointInRange(info.position*mg, aX, aY):
                ret = i
                if not aQualifier & c4d.QUALIFIER_SHIFT: break

        return ret


    def MoveHandle(self, aObj, aUndo, aMouse_pos, aHit_id, aQualifier, aBd):
        
        if hivemind.Borg().fatalError: return True

        mg = aObj.GetUpMg() * aUndo.GetMl()

        info = c4d.HandleInfo()
        self.GetHandle(aObj, aHit_id, info)

        self.SetHandle(aObj, aHit_id, info.CalculateNewPosition(aBd, mg, aMouse_pos), info)

        return True


    def GetVirtualObjects(self, aObj, aHierarchyhelp):
                
        if hivemind.Borg().fatalError: return
        
        if self.selftest:
            return self.generateStairs(aObj)
        
        dirtychild = False

        if smart.checkIsType(aObj.GetDown(), ids.PLUGIN_ID):
            if not self.childHack:
                self.childHack = True
                dirtychild = True
            else: 
                child = aObj.GetDown()
                while child:
                    if child.IsDirty(c4d.DIRTY_DATA | c4d.DIRTY_CHILDREN | c4d.DIRTY_MATRIX):
                        dirtychild = True
                        break
                    if smart.checkIsType(child.GetDown(), ids.PLUGIN_ID):
                        child = child.GetDown()
                    else:
                        break
        else:
            if self.childHack: 
                self.childHack = False
                dirtychild = True
        
        if not dirtychild:        
            if smart.checkIsType(aObj.GetUp(), ids.PLUGIN_ID):
                if not self.parentHack:
                    self.parentHack = True
                    dirtychild = True
                else: 
                    parent = aObj.GetUp()
                    while parent:
                        if parent.IsDirty(c4d.DIRTY_DATA | c4d.DIRTY_CHILDREN | c4d.DIRTY_MATRIX):
                            dirtychild = True
                            break
                        if smart.checkIsType(parent.GetUp(), ids.PLUGIN_ID):
                            parent = parent.GetUp()
                        else:
                            break
            else:
                if self.parentHack: 
                    self.parentHack = False
                    dirtychild = True

        dirty = aObj.CheckCache(aHierarchyhelp) or aObj.IsDirty(c4d.DIRTY_DATA | c4d.DIRTY_CHILDREN | c4d.DIRTY_MATRIX) or self.linkedHack or dirtychild or self.linkedHack2 or self.linkedHack3
        
        if dirty:
            if smart.checkIsType(aObj.GetDown(), ids.PLUGIN_ID):
                child = aObj.GetDown()
                while child:
                    child.SetDirty(c4d.DIRTY_DATA)
                    if smart.checkIsType(child.GetDown(), ids.PLUGIN_ID):
                        child = child.GetDown()
                    else:
                        break
                    
        if dirty is False: return aObj.GetCache(aHierarchyhelp)
        
        return self.generateStairs(aObj)


    def GetDEnabling(self, aNode, aId, aT_data, aFlags, aItemdesc):
        
        if hivemind.Borg().fatalError: return False
                        
        stair = stairobj.Stair(aNode)
        
        if aId[0].id == ids.SMARTSTAIRS_TYPE_INHERIT:
            return stair.is_child
        
        if aId[0].id == ids.SMARTSTAIRS_TREADS_INHERIT:
            return stair.is_child
            
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_WIDTH:
            return not (stair.is_child and stair.parent.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT)
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HEIGHT:
            return not stair.use_step_dims
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_DEPTH:
            return not stair.use_step_dims and not stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_SHEIGHT:
            return stair.use_step_dims
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_SDEPTH:
            return stair.use_step_dims and not stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
                
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_ANCHOR:
            return not stair.is_child and not stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
                
        base_width_override = (stair.is_child and 
                               stair.parent.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT and
                               stair.parent.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE)
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_PASTEBTN:
            return hivemind.Borg().copy is not None and hivemind.Borg().copy != aNode
        
        inherit_type = False
        if stair.is_child and stair.inherit_type:
            inherit_type = True   
            
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_TYPE:
            return not inherit_type
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_ZIGZAG:
            return (stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_STD or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3) and not inherit_type
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_TOBOTTOM:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_BOTTOMOFFSET:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HELIX_DEG:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX and not inherit_type
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HELIX_INRAD:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX and not inherit_type
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HELIX_HASCYL:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX and not inherit_type
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HELIX_CYLRAD:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX and not inherit_type
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_CHILDBTN:
            return stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX and not stair.is_parent
        
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_BASEWIDTH:
            return ((stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or 
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3) and not
                    base_width_override) and not inherit_type

        if aId[0].id == ids.SMARTSTAIRS_STAIRS_BASEHEIGHT:
            return (stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_STD or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE or 
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or
                    stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3)  and not inherit_type

        if aId[0].id == ids.SMARTSTAIRS_STAIRS_BASEDISTANCE:
            return stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE  and not inherit_type
        
        inherit_treadss = False
        steps = stair.has_treads or stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS
        if stair.is_child and stair.inherit_treads:
            steps = False
            inherit_treadss = True
            
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HASTREADS:
            return (not inherit_treadss and not stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS and 
                    (stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_STD or stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID))
        
        if aId[0].id == ids.SMARTSTAIRS_TREADS_HEIGHT:
            return steps

        if aId[0].id == ids.SMARTSTAIRS_TREADS_DEPTH:
            return steps

        if aId[0].id == ids.SMARTSTAIRS_TREADS_OFFSETZ:
            return steps and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
        
        if aId[0].id == ids.SMARTSTAIRS_TREADS_OFFSETY:
            return steps and (stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or
                              stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3)

        if aId[0].id == ids.SMARTSTAIRS_TREADS_HASFILLET:
            return steps

        if aId[0].id == ids.SMARTSTAIRS_TREADS_FILLETRAD:
            return steps and stair.treads_have_fillet

        if aId[0].id == ids.SMARTSTAIRS_TREADS_FILLETSUB:
            return steps and stair.treads_have_fillet

        if aId[0].id == ids.SMARTSTAIRS_TREADS_LINK:
            return steps and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX

        if aId[0].id == ids.SMARTSTAIRS_TREADS_LINKSCALE:
            return steps and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
        
        inherit_railing = False
        railing = stair.has_railing
        if stair.is_child and stair.inherit_railing:
            railing = False
            inherit_railing = True
            
        if aId[0].id == ids.SMARTSTAIRS_RAILING_INHERIT:
            return stair.is_child
            
        if aId[0].id == ids.SMARTSTAIRS_STAIRS_HASRAILING:
            return not inherit_railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_TYPE:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_HEIGHT:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_RAD1:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_RAD2:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_OFFSETX:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_OFFSETZ:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_OVERTOP:
            return railing and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_OVERBOTTOM:
            return railing and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_NTH:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_LINK:
            return railing
        
        if aId[0].id == ids.SMARTSTAIRS_RAILING_LINKSCALE:
            return railing        

        if aId[0].id == ids.SMARTSTAIRS_PLATFORM_DIRECTION:
            return stair.is_parent

        if aId[0].id == ids.SMARTSTAIRS_PLATFORM_WIDTH:
            return stair.is_parent and not stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT

        if aId[0].id == ids.SMARTSTAIRS_PLATFORM_HEIGHT:
            return stair.is_parent

        if aId[0].id == ids.SMARTSTAIRS_PLATFORM_DEPTH:
            return stair.is_parent
        
        return True


    def generateStairs(self, aObj):
        
        self.linkedHack = False
        self.linkedHack2 = False
        self.linkedHack3 = False

        # check param logic and limits
        self.checkAndSetLimits(aObj)
        
        stair = stairobj.Stair(aObj)
        
        if self.selftest and not stair.is_child:
            stair.setRandom(self.testtype)

        stairsNull = None
        
        # build stuff
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_STD:
            stairsNull = worker.buildStandardStairs(stair)
        
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID:
            stairsNull = worker.buildSolidStairs(stair)

        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE:
            stairsNull = worker.buildSingleStairs(stair)

        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE:
            stairsNull = worker.buildDoubleStairs(stair)
            
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2:
            stairsNull = worker.buildDoubleStairsOuter(stair)
            
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3:
            stairsNull = worker.buildDoubleStairsOuterRamp(stair)
            
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
            stairsNull = worker.buildStairsHelix(stair)
        
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS:
            stairsNull = worker.buildStepsOnlyStairs(stair)
                        
        self.linkedHack = stair.treads_link is not None            
        self.linkedHack2 = stair.railing_link is not None            
        self.linkedHack3 = stair.railing_hlink is not None  
        
        platform = None          
        
        if stair.reverse:
            stairsNull.SetAbsRot(c4d.Vector(3.141, 0, 0))
        
        if stair.is_parent:
            child_stair = stairobj.Stair(aObj.GetDown())
            
            if self.linkedHack:
                child_stair.obj.SetDirty(c4d.DIRTY_DATA)
                
            if self.linkedHack2:
                child_stair.obj.SetDirty(c4d.DIRTY_DATA)
                
            if self.linkedHack3:
                child_stair.obj.SetDirty(c4d.DIRTY_DATA)

            posy = stair.height# - stair.treads_height
            
            if stair.reverse:
                posy = stair.step_dims.y# - stair.treads_height

            if child_stair.reverse:
                posy = - (child_stair.height - stair.height)
                
            if stair.reverse and child_stair.reverse:
                posy = - child_stair.height + stair.step_dims.y# - stair.treads_height# + child_stair.treads_height

            if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT:
                posz = stair.depth / 2.0 + child_stair.depth / 2.0 + stair.platform_dims.z            
                child_stair.obj.SetAbsPos(c4d.Vector(0.0, posy, posz))
                child_stair.obj.SetAbsRot(c4d.Vector(0.0, 0.0, 0.0))
                child_stair.width = stair.width
                
                platform, platform_dims = worker.buildPlatform(stairsNull, stair, child_stair)                
                
            if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_LEFT:
                posx = - stair.width / 2.0 - child_stair.depth / 2.0 - stair.platform_width
                posz = stair.depth / 2.0 + child_stair.width / 2.0 + stair.platform_dims.z
                
                platform, platform_dims = worker.buildPlatform(stairsNull, stair, child_stair)
                
                child_stair.obj.SetAbsPos(c4d.Vector(posx, posy, posz))
                child_stair.obj.SetAbsRot(c4d.Vector(math.pi / 2.0, 0.0, 0.0))                
                
            if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_RIGHT:
                posx = stair.width / 2.0 + child_stair.depth / 2.0 + stair.platform_width
                posz = stair.depth / 2.0 + child_stair.width / 2.0 + stair.platform_dims.z
                
                platform, platform_dims = worker.buildPlatform(stairsNull, stair, child_stair)
                
                if platform is not None:
                    ppos = platform.GetAbsPos() 
                    if stair.reverse:
                        ppos.x -= stair.platform_width
                    else:
                        ppos.x += stair.platform_width                    
                    platform.SetAbsPos(ppos)
                    
                child_stair.obj.SetAbsPos(c4d.Vector(posx, posy, posz))
                child_stair.obj.SetAbsRot(c4d.Vector(- math.pi / 2.0, 0.0, 0.0))                
                
            if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT:
                posx = - stair.width / 2.0 - child_stair.width / 2.0 - stair.platform_width
                posz = stair.depth / 2.0 - child_stair.depth / 2.0 - stair.step_dims.z
                
                if child_stair.reverse:
                    posz += child_stair.step_dims.z
                    
                platform, platform_dims = worker.buildPlatform(stairsNull, stair, child_stair)
                
                if platform is not None:
                    ppos = platform.GetAbsPos() 
                    if stair.reverse:
                        ppos.x = - stair.width / 2.0 + platform_dims.x / 2.0
                    else:
                        ppos.x = stair.width / 2.0 - platform_dims.x / 2.0
                    platform.SetAbsPos(ppos)
                    
                child_stair.obj.SetAbsPos(c4d.Vector(posx, posy, posz))
                child_stair.obj.SetAbsRot(c4d.Vector(- math.pi, 0.0, 0.0))                               
                
            if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT:
                posx = stair.width / 2.0 + child_stair.width / 2.0 + stair.platform_width
                posz = stair.depth / 2.0 - child_stair.depth / 2.0 - stair.step_dims.z
                
                if child_stair.reverse:
                    posz += child_stair.step_dims.z
                
                platform, platform_dims = worker.buildPlatform(stairsNull, stair, child_stair)
                
                if platform is not None:                    
                    ppos = platform.GetAbsPos() 
                    if stair.reverse:
                        ppos.x = stair.width / 2.0 - platform_dims.x / 2.0
                    else:
                        ppos.x = - stair.width / 2.0 + platform_dims.x / 2.0
                    platform.SetAbsPos(ppos)
                    
                child_stair.obj.SetAbsPos(c4d.Vector(posx, posy, posz))
                child_stair.obj.SetAbsRot(c4d.Vector(- math.pi, 0.0, 0.0))                
                
            if platform is not None:
                self.platform_bb = platform_dims
                self.platform_pos = platform.GetAbsPos()
                
                if (stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or
                    stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT):
                    self.platform_bb.z += stair.step_dims.z
                    self.platform_pos.z -= stair.step_dims.z
                    
                if stair.reverse:
                    self.platform_pos.z += stair.depth
                    self.platform_pos.x *= -1
                    
                self.platform_pos.z += stair.depth / 2.0

        if stairsNull is not None:
            if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
                stairsNull.SetAbsPos(c4d.Vector(0.0, 0.0, stair.depth / 2.0))
            if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                stairsNull.SetAbsPos(c4d.Vector(0.0, 0.0, - stair.depth / 2.0))
                
            if stair.is_parent:
                pos = child_stair.obj.GetAbsPos()
                if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
                    pos.z += stair.depth / 2.0
                if stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                    pos.z -= stair.depth / 2.0
                    
                if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT:
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
                        pos.z -= child_stair.depth / 2.0
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                        pos.z += child_stair.depth / 2.0
                        
                if (stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or 
                    stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT):
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
                        pos.z += child_stair.depth / 2.0
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                        pos.z -= child_stair.depth / 2.0
                        
                if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_LEFT:
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
                        pos.x += child_stair.depth / 2.0
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                        pos.x -= child_stair.depth / 2.0
                        
                if stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_RIGHT:
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT:
                        pos.x -= child_stair.depth / 2.0
                    if child_stair.stair_anchor == ids.SMARTSTAIRS_STAIRS_ANCHOR_BACK:
                        pos.x += child_stair.depth / 2.0
                    
                child_stair.obj.SetAbsPos(pos)
                
            stairsNull.Message(c4d.MSG_UPDATE)

            return stairsNull


    def checkAndSetLimits(self, aObj):

        stair = stairobj.Stair(aObj)

        if stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_STD and stair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_SOLID:
            stair.has_treads = True
        
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE:               
            if stair.base_width > stair.width:
                stair.base_width = stair.width        
                
        if stair.is_parent:
            stair.child.stair_anchor = ids.SMARTSTAIRS_STAIRS_ANCHOR_FRONT
                
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
            stair.depth = stair.width
            stair.stair_anchor = ids.SMARTSTAIRS_STAIRS_ANCHOR_CENTER

        if (stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE or 
            stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or
            stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3):                
            if stair.base_distance > stair.width / 2.0:
                stair.base_distance = stair.width / 2.0
                
            if stair.base_width >  stair.width / 2.0 - stair.base_distance:
                stair.base_width = stair.width / 2.0 - stair.base_distance
                        
        if (stair.is_parent and stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT and
            stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE and 
            stair.child.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE):
            stair.child.base_width = stair.base_width
                        
        if stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS and not stair.has_treads:
            stair.has_treads = True
            
        if stair.is_parent and stair.child.step_dims.z > stair.step_dims.z:
            if (stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or
                stair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT) and stair.child.reverse:
                    off = stair.child.step_dims.z - stair.step_dims.z
                    if stair.platform_depth < off:
                        stair.platform_depth = off

        # the top-step may not be higher than the step height 
        if stair.treads_height > stair.step_dims.y:
            stair.treads_height = stair.step_dims.y - 0.01

        # the top-step x offset may not be bigger than the step depth
        if stair.treads_offset_z > stair.step_dims.z:
            stair.treads_offset_z = stair.step_dims.z
            
        # the top-step y offset may not be bigger than the step height
        if stair.treads_offset_y > stair.step_dims.y - stair.treads_height :
            stair.treads_offset_y = stair.step_dims.y - stair.treads_height 

        # the top step filet radius may not be bigger than half the step's height
        if stair.treads_fillet_rad > stair.treads_height / 2.0:
            stair.treads_fillet_rad = stair.treads_height / 2.0
                        
        # inner radius of helix may not be bigger than outer:
        if stair.stair_helix_innerrad > stair.width / 2.0:
            stair.stair_helix_innerrad = stair.width / 2.0    
            
        if stair.has_railing and stair.num_steps > 1:
            if stair.stair_helix_innerrad + stair.railing_offsetx * 2 + stair.railing_rad1 * 2 > stair.width / 2.0:
                stair.stair_helix_innerrad = stair.width / 2.0  - stair.railing_offsetx * 2 - stair.railing_rad1 * 2
            
        # transfer inheritance:
        if stair.is_parent and stair.child.inherit_type:
            stair.child.stairtype = stair.stairtype
            stair.child.base_width = stair.base_width
            stair.child.base_height = stair.base_height
            stair.child.base_distance = stair.base_distance
            stair.child.stair_zigzag = stair.stair_zigzag
            stair.child.stair_tobottom = stair.stair_tobottom
            stair.child.stair_bottomoffset = stair.stair_bottomoffset
            stair.child.stair_helix_deg = stair.stair_helix_deg
            stair.child.stair_helix_innerrad = stair.stair_helix_innerrad
            stair.child.stair_helix_hascyl = stair.stair_helix_hascyl
            stair.child.stair_helix_cylrad = stair.stair_helix_cylrad
        
        if stair.is_parent and stair.child.inherit_treads:
            stair.child.has_treads = stair.has_treads
            stair.child.treads_height = stair.treads_height
            stair.child.treads_depth = stair.treads_depth
            stair.child.treads_offset_z = stair.treads_offset_z
            stair.child.treads_offset_y = stair.treads_offset_y
            stair.child.treads_have_fillet = stair.treads_have_fillet
            stair.child.treads_fillet_rad = stair.treads_fillet_rad
            stair.child.treads_fillet_sub = stair.treads_fillet_sub
            if stair.child.treads_link is not None and stair.treads_link is not None:
                stair.child.treads_link = stair.treads_link
            stair.child.treads_link_scale = stair.treads_link_scale
            
        if stair.is_parent and stair.child.inherit_railing:
            stair.child.has_railing = stair.has_railing
            stair.child.railing_type = stair.railing_type
            stair.child.railing_height = stair.railing_height
            stair.child.railing_rad1 = stair.railing_rad1
            stair.child.railing_rad2 = stair.railing_rad2
            stair.child.railing_offsetx = stair.railing_offsetx
            stair.child.railing_offsetz = stair.railing_offsetz
            stair.child.railing_overtop = stair.railing_overtop
            stair.child.railing_overbottom = stair.railing_overbottom
            stair.child.railing_nth = stair.railing_nth
            if stair.child.railing_link is not None and stair.railing_link is not None:
                stair.child.railing_link = stair.railing_link
            stair.child.railing_link_scale = stair.railing_link_scale
            if stair.child.railing_hlink is not None and stair.railing_hlink is not None:
                stair.child.railing_hlink = stair.railing_hlink
            stair.child.railing_hlink_scale = stair.railing_hlink_scale
            
        if stair.is_parent and stair.child.inherit_materials:
            if stair.child.mat_linkbase is not None and stair.mat_linkbase is not None:
                stair.child.mat_linkbase = stair.mat_linkbase
            if stair.child.mat_linkstep is not None and stair.mat_linkstep is not None:
                stair.child.mat_linkstep = stair.mat_linkstep
            if stair.child.mat_linkplatform is not None and stair.mat_linkplatform is not None:
                stair.child.mat_linkplatform = stair.mat_linkplatform
            if stair.child.mat_linkrailing is not None and stair.mat_linkrailing is not None:
                stair.child.mat_linkrailing = stair.mat_linkrailing
            if stair.child.mat_handrail is not None and stair.mat_handrail is not None:
                stair.child.mat_handrail = stair.mat_handrail
        
        
    def getParameters(self, aObj):
        
        if hivemind.Borg().copy is None: return
        
        stair = stairobj.Stair(aObj)
        source = stairobj.Stair(hivemind.Borg().copy)
        
        stair.width = source.width
        stair.height = source.height
        stair.depth = source.depth
        stair.use_step_dims = source.use_step_dims
        stair.num_steps = source.num_steps
        stair.reverse = source.reverse
        stair.stair_anchor = source.stair_anchor

        stair.inherit_type = source.inherit_type
        stair.stairtype = source.stairtype
        stair.base_width = source.base_width
        stair.base_height = source.base_height
        stair.base_distance = source.base_distance
        stair.stair_zigzag = source.stair_zigzag
        stair.stair_tobottom = source.stair_tobottom
        stair.stair_bottomoffset = source.stair_bottomoffset
        stair.stair_helix_deg = source.stair_helix_deg
        stair.stair_helix_innerrad = source.stair_helix_innerrad
        stair.stair_helix_hascyl = source.stair_helix_hascyl
        stair.stair_helix_cylrad = source.stair_helix_cylrad
        
        stair.inherit_treads = source.inherit_treads
        stair.has_treads = source.has_treads
        stair.treads_height = source.treads_height
        stair.treads_depth = source.treads_depth
        stair.treads_offset_z = source.treads_offset_z
        stair.treads_offset_y = source.treads_offset_y
        stair.treads_have_fillet = source.treads_have_fillet
        stair.treads_fillet_rad = source.treads_fillet_rad
        stair.treads_fillet_sub = source.treads_fillet_sub
        if stair.treads_link is not None and source.treads_link is not None:
            stair.treads_link = source.treads_link
        stair.treads_link_scale = source.treads_link_scale
            
        stair.inherit_railing = source.inherit_railing
        stair.has_railing = source.has_railing
        stair.railing_type = source.railing_type
        stair.railing_height = source.railing_height
        stair.railing_rad1 = source.railing_rad1
        stair.railing_rad2 = source.railing_rad2
        stair.railing_offsetx = source.railing_offsetx
        stair.railing_offsetz = source.railing_offsetz
        stair.railing_overtop = source.railing_overtop
        stair.railing_overbottom = source.railing_overbottom
        stair.railing_nth = source.railing_nth
        if stair.railing_link is not None and source.railing_link is not None:
            stair.railing_link = source.railing_link
        stair.railing_link_scale = source.railing_link_scale
        if stair.railing_hlink is not None and source.railing_hlink is not None:
            stair.railing_hlink = source.railing_hlink
        stair.railing_hlink_scale = source.railing_hlink_scale
            
        stair.inherit_materials = source.inherit_materials
        if stair.mat_linkbase is not None and source.mat_linkbase is not None:
            stair.mat_linkbase = source.mat_linkbase
        if stair.mat_linkstep is not None and source.mat_linkstep is not None:
            stair.mat_linkstep = source.mat_linkstep
        if stair.mat_linkplatform is not None and source.mat_linkplatform is not None:
            stair.mat_linkplatform = source.mat_linkplatform
        if stair.mat_linkrailing is not None and source.mat_linkrailing is not None:
            stair.mat_linkrailing = source.mat_linkrailing
        if stair.mat_handrail is not None and source.mat_handrail is not None:
            stair.mat_handrail = source.mat_handrail
            
        stair.platform_direction = source.platform_direction
        stair.platform_width = source.platform_width
        stair.platform_depth = source.platform_depth
        