# -*- coding: utf-8 -*-

import c4d

from smartstairspackage import ids


def generateStairSpline(aStair):
    
    if not aStair.stair_zigzag: 
        numPointsAdditional = 1
        numPoints = aStair.num_steps * 3 + numPointsAdditional
    else:
        numPointsAdditional = 0
        numPoints = aStair.num_steps * 4 + numPointsAdditional
        
    s = c4d.SplineObject(numPoints, c4d.SPLINETYPE_LINEAR)
    
    treads_height = aStair.treads_height
    
    if aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2:
        treads_height = 0
        
    #zigzag top
    cnt = 0
    for i in range(0, aStair.num_steps):
        ypos = aStair.step_dims.y * i - treads_height
        if ypos < 0: ypos = 0
        
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * i))
        cnt += 1
        
        ypos = aStair.step_dims.y * (i + 1) - treads_height
        if ypos < 0: ypos = 0
        
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * i))
        cnt += 1
        
    # close top right
    s.SetPoint(cnt, c4d.Vector(0.0, aStair.height - treads_height, aStair.depth))
    cnt += 1
    
    if aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 and aStair.is_parent and not aStair.reverse:
        s.SetPoint(aStair.num_steps*2-1, c4d.Vector(0.0, aStair.height - aStair.treads_height, aStair.depth - aStair.step_dims.z))
        s.SetPoint(aStair.num_steps*2, c4d.Vector(0.0, aStair.height - aStair.treads_height, aStair.depth))
        
    if aStair.is_child and aStair.reverse:
        s.SetPoint(aStair.num_steps*2-1, c4d.Vector(0.0, aStair.height - aStair.parent.treads_height, aStair.depth - aStair.step_dims.z))
        s.SetPoint(aStair.num_steps*2, c4d.Vector(0.0, aStair.height - aStair.parent.treads_height, aStair.depth))
        
    if aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 and aStair.is_parent and aStair.reverse:
        s.SetPoint(1, c4d.Vector(0.0, aStair.step_dims.y - aStair.treads_height, 0))
        s.SetPoint(2, c4d.Vector(0.0, aStair.step_dims.y - aStair.treads_height, aStair.step_dims.z))
      
    if not aStair.stair_zigzag:  
        # step one down
        ypos = aStair.height - aStair.step_dims.y - (aStair.base_height + aStair.treads_height)
        if ypos < 0: ypos = 0
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.depth))
        cnt += 1
        
        for i in range(0, aStair.num_steps - 1):
            zpos = aStair.depth - aStair.step_dims.z * (i + 1)
            ypos = aStair.height - aStair.step_dims.y * (i + 2) - (aStair.base_height + aStair.treads_height)
            if ypos < 0:
                zpos -= ypos*(aStair.step_dims.z/aStair.step_dims.y)
                ypos = 0 
                
            if zpos > aStair.depth:
                zpos = aStair.depth
            
            s.SetPoint(cnt, c4d.Vector(0.0, ypos, zpos))
            cnt += 1
    else:
        # step one down
        ypos = aStair.height - aStair.step_dims.y * 2.0 - aStair.base_height 
        if aStair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2:
            ypos -= aStair.treads_height
        if ypos < 0: ypos = 0
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.depth))
        cnt += 1
        
        #zigzag bottom
        for i in range(0, aStair.num_steps - 1):
            num = aStair.num_steps - 1
            ypos = aStair.step_dims.y * (num - (i + 1)) - treads_height - aStair.base_height
            if ypos < 0: ypos = 0
            
            s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * (num - i)))
            cnt += 1
            
            ypos = aStair.step_dims.y * (num - (i + 2)) - treads_height - aStair.base_height
            if ypos < 0: ypos = 0
            
            s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * (num - i)))
            cnt += 1
            
    s[c4d.SPLINEOBJECT_CLOSED] = True
    s.SetAbsPos(c4d.Vector(- aStair.width / 2.0, 0.0, - aStair.depth / 2.0))
    
    return s


def generateRampSpline(aStair):
    
    if not aStair.stair_zigzag: 
        numPointsAdditional = 2 if not aStair.is_parent else 3
        numPoints = aStair.num_steps * 2 + numPointsAdditional
    else:
        numPointsAdditional = 0 if not aStair.is_parent else 1
        numPoints = aStair.num_steps * 4 + numPointsAdditional
        
    if (aStair.is_parent and not aStair.reverse) or (aStair.is_child and aStair.reverse):
        numPoints += 1

    s = c4d.SplineObject(numPoints, c4d.SPLINETYPE_LINEAR)

    cnt = 0
            
    s.SetPoint(cnt, c4d.Vector(0.0, 0.0, 0.0))
    cnt += 1
        
    #top
    for i in range(0, aStair.num_steps):       
        ypos = aStair.step_dims.y * (i + 1)
        if ypos < 0: ypos = 0
        
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * i - aStair.treads_offset_z))
        cnt += 1    
    
    # close top right    
    if (aStair.is_parent and not aStair.reverse) or (aStair.is_child and aStair.reverse):
        s.SetPoint(cnt, c4d.Vector(0.0, aStair.height - aStair.treads_height, aStair.depth - aStair.step_dims.z))
        cnt += 1
        
        s.SetPoint(cnt, c4d.Vector(0.0, aStair.height - aStair.treads_height, aStair.depth))
        cnt += 1
    else:
        s.SetPoint(cnt, c4d.Vector(0.0, aStair.height, aStair.depth))
        cnt += 1
    
    if not aStair.stair_zigzag:      
        # step one down
        ypos = aStair.height - aStair.step_dims.y - (aStair.base_height + aStair.treads_height)
        if ypos < 0: ypos = 0
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.depth))
        cnt += 1
        
        for i in range(0, aStair.num_steps - 1):
            zpos = aStair.depth - aStair.step_dims.z * (i + 1)
            ypos = aStair.height - aStair.step_dims.y * (i + 2) - (aStair.base_height + aStair.treads_height)
            if ypos < 0:
                zpos -= ypos*(aStair.step_dims.z/aStair.step_dims.y)
                ypos = 0 
                
            if zpos > aStair.depth:
                zpos = aStair.depth
            
            s.SetPoint(cnt, c4d.Vector(0.0, ypos, zpos))
            cnt += 1
    else:
        # step one down
        ypos = aStair.height - aStair.step_dims.y * 2.0 - aStair.base_height
        if ypos < 0: ypos = 0
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.depth))
        cnt += 1
        
        #zigzag bottom
        for i in range(0, aStair.num_steps - 1):
            num = aStair.num_steps - 1
            ypos = aStair.step_dims.y * (num - (i + 1)) - aStair.base_height
            if ypos < 0: ypos = 0
            
            s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * (num - i)))
            cnt += 1
            
            ypos = aStair.step_dims.y * (num - (i + 2)) - aStair.base_height
            if ypos < 0: ypos = 0
            
            s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * (num - i)))
            cnt += 1

    
    s[c4d.SPLINEOBJECT_CLOSED] = True
    s.SetAbsPos(c4d.Vector(- aStair.width / 2.0, 0.0, - aStair.depth / 2.0))
    
    return s


def generateStairSpline_Solid(aStair):

    numPointsAdditional = 2
    numPoints = aStair.num_steps * 2 + numPointsAdditional

    s = c4d.SplineObject(numPoints, c4d.SPLINETYPE_LINEAR)
    
    #zigzag top
    cnt = 0
    for i in range(0, aStair.num_steps):
        ypos = aStair.step_dims.y * i - aStair.treads_height 
        if ypos < 0: ypos = 0
        
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * i))
        cnt += 1
        
        ypos = aStair.step_dims.y * (i + 1) - aStair.treads_height
        if ypos < 0: ypos = 0
        
        s.SetPoint(cnt, c4d.Vector(0.0, ypos, aStair.step_dims.z * i))
        cnt += 1
        
    # close top right
    s.SetPoint(cnt, c4d.Vector(0.0, aStair.height - aStair.treads_height, aStair.depth))
    cnt += 1
    
    if aStair.is_child and aStair.reverse:
        s.SetPoint(aStair.num_steps*2-1, c4d.Vector(0.0, aStair.height - aStair.parent.treads_height, aStair.depth - aStair.step_dims.z))
        s.SetPoint(aStair.num_steps*2, c4d.Vector(0.0, aStair.height - aStair.parent.treads_height, aStair.depth))
        
    if aStair.is_child and aStair.reverse and aStair.parent.reverse:
        s.SetPoint(aStair.num_steps*2-1, c4d.Vector(0.0, aStair.height, aStair.depth - aStair.step_dims.z))
        s.SetPoint(aStair.num_steps*2, c4d.Vector(0.0, aStair.height, aStair.depth))
    # step all the way down
    s.SetPoint(cnt, c4d.Vector(0.0, 0.0, aStair.depth))
    
    if aStair.stair_tobottom:
        ground = 0
        mo = aStair.obj.GetMg().off
        if aStair.stair_bottomoffset < mo.y:
            ground = mo.y - aStair.stair_bottomoffset
            
            s.SetPoint(0, c4d.Vector(0.0, - ground, 0.0))
            s.SetPoint(cnt, c4d.Vector(0.0, - ground, aStair.depth))
    
    s[c4d.SPLINEOBJECT_CLOSED] = True
    s.SetAbsPos(c4d.Vector(- aStair.width / 2.0, 0.0, - aStair.depth / 2.0))
    
    return s


def generateWedge(offset, depth, height, quad = False):
    
    numPoints = 4
    
    s = c4d.SplineObject(numPoints, c4d.SPLINETYPE_LINEAR)
    
    s.SetPoint(0, c4d.Vector(0.0, 0.0, 0.0))
    s.SetPoint(1, c4d.Vector(0.0, 0.0, depth + offset))
    s.SetPoint(2, c4d.Vector(0.0, - height, offset + (depth if quad else 0)))    
    s.SetPoint(3, c4d.Vector(0.0, - height, 0.0))    
    
    s[c4d.SPLINEOBJECT_CLOSED] = True
    
    return s


def generateTreadWedge(aStair):
    
    numPoints = 4
    
    inner = aStair.stair_helix_innerrad
    outer = aStair.width / 2.0
    
    if aStair.treads_have_fillet:
        outer -= aStair.treads_fillet_rad
    
    num = aStair.num_steps
    deg = abs(aStair.stair_helix_deg) / float(num)
    
    ci = (deg * inner + aStair.treads_depth) / 2.0 - aStair.treads_fillet_rad
    co = (deg * outer + aStair.treads_depth) / 2.0 - aStair.treads_fillet_rad
    
    if ci < 0: ci = 0
    
    l = outer - inner + aStair.stair_helix_innerrad
    
    s = c4d.SplineObject(numPoints, c4d.SPLINETYPE_LINEAR)
    
    s.SetPoint(0, c4d.Vector(ci, 0.0, aStair.stair_helix_innerrad))
    s.SetPoint(1, c4d.Vector(co, 0.0, l))
    s.SetPoint(2, c4d.Vector(- co, 0.0, l))    
    s.SetPoint(3, c4d.Vector(- ci, 0.0, aStair.stair_helix_innerrad))    
    
    s[c4d.SPLINEOBJECT_CLOSED] = True
    
    return s