# -*- coding: utf-8 -*-

import math

import c4d
from c4d import plugins

from smartstairspackage import splinegen

from smartstairspackage import ids


def buildStandardStairs(aStair):

    null = getNull()
    extrude = getExtrude(aStair.width, null, aStair.mat_linkbase)
    getSpline(splinegen.generateStairSpline(aStair), extrude)

    if aStair.has_treads:
        stepInstance = getStepInstance(aStair, null)
        getStepInstances(aStair, stepInstance, null)
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildSolidStairs(aStair):

    null = getNull()
    extrude = getExtrude(aStair.width, null, aStair.mat_linkbase)
    getSpline(splinegen.generateStairSpline_Solid(aStair), extrude)

    if aStair.has_treads:
        stepInstance = getStepInstance(aStair, null)
        getStepInstances(aStair, stepInstance, null)
    
    if aStair.has_railing and aStair.num_steps > 1:    
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildSingleStairs(aStair):

    null = getNull()
    extrude = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    getSpline(splinegen.generateStairSpline(aStair), extrude)

    if aStair.has_treads:
        stepInstance = getStepInstance(aStair, null)
        getStepInstances(aStair, stepInstance, null)
        
    extrude.SetAbsPos(c4d.Vector(aStair.width / 2.0 - aStair.base_width / 2.0, 0.0, 0.0))
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildDoubleStairs(aStair):

    null = getNull()
    extrude = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    extrude2 = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    spline = getSpline(splinegen.generateStairSpline(aStair), extrude)

    spline2 = spline.GetClone()
    spline2.InsertUnder(extrude2)

    if aStair.has_treads:
        stepInstance = getStepInstance(aStair, null)
        getStepInstances(aStair, stepInstance, null)

    extrude.SetAbsPos(c4d.Vector(aStair.base_distance, 0.0, 0.0))
    extrude2.SetAbsPos(c4d.Vector(aStair.width - aStair.base_width - aStair.base_distance, 0.0, 0.0))
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildDoubleStairsOuter(aStair):

    null = getNull()
    extrude = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    extrude2 = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    spline = getSpline(splinegen.generateStairSpline(aStair), extrude)

    spline2 = spline.GetClone()
    spline2.InsertUnder(extrude2)

    if aStair.has_treads:
        stepInstance = getStepInstance(aStair, null)
        getStepInstances(aStair, stepInstance, null)

    extrude.SetAbsPos(c4d.Vector(0.0, 0.0, 0.0))
    extrude2.SetAbsPos(c4d.Vector(aStair.width - aStair.base_width, 0.0, 0.0))
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildDoubleStairsOuterRamp(aStair):

    null = getNull()
    extrude = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    extrude2 = getExtrude(aStair.base_width, null, aStair.mat_linkbase)
    spline = getSpline(splinegen.generateRampSpline(aStair), extrude)

    spline2 = spline.GetClone()
    spline2.InsertUnder(extrude2)

    if aStair.has_treads:
        stepInstance = getStepInstance(aStair, null)
        getStepInstances(aStair, stepInstance, null)

    extrude.SetAbsPos(c4d.Vector(0.0, 0.0, 0.0))
    extrude2.SetAbsPos(c4d.Vector(aStair.width - aStair.base_width, 0.0, 0.0))
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildStairsHelix(aStair):
    
    null = getNull()    
    nullReference = getReference(null)
    
    extrude = getExtrude(aStair.treads_height, null, aStair.mat_linkstep, True)
    getSpline(splinegen.generateTreadWedge(aStair), extrude)
    
    if aStair.treads_have_fillet:
        extrude[c4d.CAP_START] = 3
        extrude[c4d.CAP_STARTSTEPS] = aStair.treads_fillet_sub
        extrude[c4d.CAP_STARTRADIUS] = aStair.treads_fillet_rad
        extrude[c4d.CAP_END] = 3
        extrude[c4d.CAP_ENDSTEPS] = aStair.treads_fillet_sub
        extrude[c4d.CAP_ENDRADIUS] = aStair.treads_fillet_rad
        extrude[c4d.EXTRUDEOBJECT_MOVE] = c4d.Vector(0.0, aStair.treads_height - aStair.treads_fillet_rad * 2, 0.0)

    extrude.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPOBJ))
    extrude.InsertUnder(nullReference)
    
    instance = c4d.BaseObject(c4d.Oinstance)
    instance[c4d.INSTANCEOBJECT_LINK] = extrude
    instance[c4d.INSTANCEOBJECT_RENDERINSTANCE_MODE] = True
    instance.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPINSTANCEOBJ))
    
    nullInstances = c4d.BaseObject(c4d.Onull)
    nullInstances.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPINSTANCESOBJ))
    nullInstances.InsertUnder(null)
    
    for i in range(aStair.num_steps):        
        obj = instance.GetClone()   
        obj.InsertUnder(nullInstances)
        
        rd = aStair.stair_helix_deg/float(aStair.num_steps)
        obj.SetAbsPos(c4d.Vector(0.0, aStair.height/float(aStair.num_steps)*i + aStair.step_dims.y - aStair.treads_height, 0))
        obj.SetAbsRot(c4d.Vector(rd*i + rd / 2.0, 0, 0))
        
        if aStair.treads_have_fillet:           
            obj.SetAbsPos(c4d.Vector(0.0, aStair.height/float(aStair.num_steps)*i + aStair.step_dims.y - aStair.treads_height + aStair.treads_fillet_rad, 0))

    if aStair.stair_helix_hascyl:
        cyl = c4d.BaseObject(c4d.Ocylinder)
        tag = cyl.MakeTag(c4d.Tphong)
        tag[c4d.PHONGTAG_PHONG_ANGLELIMIT] = True
        
        off = aStair.stair_helix_innerrad if aStair.stair_helix_innerrad > 0 else 0
        cyl[c4d.PRIM_CYLINDER_RADIUS] = aStair.stair_helix_cylrad + off
        cyl[c4d.PRIM_CYLINDER_HEIGHT] = aStair.height
        cyl.SetAbsPos(c4d.Vector(0.0, aStair.height / 2.0, 0))
        cyl.InsertUnder(null)
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)
          
    return null


def buildStepsOnlyStairs(aStair):

    null = getNull()
    stepInstance = getStepInstance(aStair, null)
    getStepInstances(aStair, stepInstance, null)  
    
    if aStair.has_railing and aStair.num_steps > 1:
        handleInstance = getHandleInstance(aStair, null)
        getHandleInstances(aStair, handleInstance, null)

    return null


def buildPlatform(aParent, aStair, aChild_stair):
        
    only_steps = aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS
    
    if only_steps and not aStair.has_treads:
        return None, None
       
    max_dims = c4d.Vector()
    
    platform_dims = c4d.Vector(aStair.platform_dims)    
    
    if (aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_LEFT or 
        aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_RIGHT):
        platform_dims.z = aChild_stair.width + aStair.platform_depth
        platform_dims.x += aStair.platform_width
        
    if (aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or 
        aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT):
        platform_dims.x += aChild_stair.width + aStair.platform_width        
    
    is_solid = aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID
                
    width = platform_dims.x
    if (((aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE and 
         aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE) or 
         aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE) and
         aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT):
        width = aChild_stair.base_width
        
    if aStair.reverse or aStair.num_steps == 1:
        platform_dims.y -= aStair.treads_height
        
    child_treads_height = aChild_stair.treads_height
    if aChild_stair.step_dims.z < child_treads_height:
        child_treads_height = aChild_stair.step_dims.z
        
    if child_treads_height > aChild_stair.step_dims.y: 
        child_treads_height = aChild_stair.step_dims.y
    
    if not aStair.reverse and aChild_stair.reverse:
        platform_dims.y = aChild_stair.step_dims.y
        if aStair.step_dims.y < aChild_stair.step_dims.y:
            platform_dims.y = aStair.step_dims.y
          
    wegde_height = platform_dims.y + aStair.treads_height
    
    wedge_offset = aChild_stair.step_dims.z  - (aChild_stair.step_dims.z / aChild_stair.step_dims.y * wegde_height)
    wedge_end = aChild_stair.step_dims.z  - wedge_offset
    
    if aChild_stair.num_steps > 1:
        wedge_offset += aChild_stair.step_dims.z * (aChild_stair.treads_height / aChild_stair.step_dims.y)
    
    both_treads = (aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS and aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_TREADS)
    both_single = (aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE and aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE)
    
    if wedge_offset < 0: 
        wedge_offset = 0
        wedge_end = aChild_stair.step_dims.z + aChild_stair.step_dims.z * (aChild_stair.treads_height / aChild_stair.step_dims.y)
            
    if is_solid and not aStair.reverse:
        platform_dims.y = aStair.height - aStair.treads_height
        
    solidoff = 0
    if is_solid and aStair.stair_tobottom:
        mo = aStair.obj.GetMg().off
        if aStair.stair_bottomoffset < mo.y:
            solidoff = mo.y - aStair.stair_bottomoffset
            platform_dims.y += solidoff

    platformNull = c4d.BaseObject(c4d.Onull)
    platformNull.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_PLATFORMOBJ))
    platformNull.InsertUnder(aParent)
    
    platform = c4d.BaseObject(c4d.Ocube)

    pw = width
    posx = 0
    if (aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or 
        aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT) and aChild_stair.reverse:
        pw = aStair.width + aStair.platform_width
        posx = platform_dims.x / 2.0 - pw / 2.0
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT:
            posx =  platform_dims.x / 2.0 - pw / 2.0 - aChild_stair.width
    
    max_dims.x = width
    max_dims.y = platform_dims.y
    max_dims.z = platform_dims.z
    
    if not both_treads:
        platform[c4d.PRIM_CUBE_LEN] = c4d.Vector(pw, platform_dims.y, platform_dims.z)
        platform.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_PLATFORMOBJ)) #
        platform.InsertUnder(platformNull)
        platform.SetAbsPos(c4d.Vector(posx, - platform_dims.y / 2.0, platform_dims.z / 2.0))
    else:
        max_dims.y = 0
    
    if (aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or 
        aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT) and aChild_stair.reverse:
            platform = c4d.BaseObject(c4d.Ocube)
            zlen = platform_dims.z - aChild_stair.step_dims.z + aStair.step_dims.z
            if zlen > 0.001:
                posx = - platform_dims.x / 2.0 + aChild_stair.width / 2.0
                if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT:
                    posx = platform_dims.x / 2.0 - aChild_stair.width / 2.0# + aChild_stair.width / 2
                    
                if not both_treads:
                    platform[c4d.PRIM_CUBE_LEN] = c4d.Vector(aChild_stair.width, platform_dims.y, zlen)
                    platform.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_PLATFORMOBJ)) # 
                    platform.InsertUnder(platformNull)
                    platform.SetAbsPos(c4d.Vector(posx, - platform_dims.y / 2.0, - aStair.step_dims.z + zlen / 2.0 + aChild_stair.step_dims.z))        
        
    if ((aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT or
        aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT) and not
        both_treads):
        pw = aChild_stair.width + aStair.platform_width
        
        if aChild_stair.reverse:
            pw = aStair.platform_width
            
        platform = c4d.BaseObject(c4d.Ocube)

        if both_single:
            pw = aChild_stair.base_width
        
        if pw > 0.001 and not (both_single and aChild_stair.reverse):
            platform[c4d.PRIM_CUBE_LEN] = c4d.Vector(pw, platform_dims.y, aStair.step_dims.z)
            platform.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_PLATFORMOBJ)  + "!")
            platform.InsertUnder(platformNull)
        
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT:
            posx = pw / 2.0 - platform_dims.x / 2.0# + aStair.width
            if aChild_stair.reverse:
                posx += aChild_stair.width
            else:
                if both_single:
                    posx += aChild_stair.width / 2.0 - pw / 2.0
                        
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT:
            posx = pw / 2.0 - platform_dims.x / 2.0 + aStair.width
            if not aChild_stair.reverse and both_single:
                posx += aChild_stair.width / 2.0 - pw / 2.0 + aStair.platform_width
            
        platform.SetAbsPos(c4d.Vector(posx, - platform_dims.y / 2.0, - aStair.step_dims.z / 2.0))      

    if aStair.has_treads:
        sdims = c4d.Vector(platform_dims.x, aStair.platform_dims.y, platform_dims.z)
        sdims.z += aStair.step_dims.z
        xoff = 0
        
        if aChild_stair.reverse and aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_FRONT:
            sdims.z += aChild_stair.step_dims.z
            
        if aChild_stair.reverse and aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_LEFT:
            sdims.x += aChild_stair.step_dims.z
            xoff = - aChild_stair.step_dims.z / 2.0
            
        if aChild_stair.reverse and aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_RIGHT:
            sdims.x += aChild_stair.step_dims.z
            xoff = aChild_stair.step_dims.z / 2.0
            
        sdims.y = aStair.treads_height
        step = c4d.BaseObject(c4d.Ocube)

        step[c4d.PRIM_CUBE_LEN] = c4d.Vector(sdims)
        step[c4d.PRIM_CUBE_DOFILLET] = aStair.treads_have_fillet
        step[c4d.PRIM_CUBE_FRAD] = aStair.treads_fillet_rad
        step[c4d.PRIM_CUBE_SUBF] = aStair.treads_fillet_sub
        step.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPOBJ))
        step.InsertUnder(platformNull)
        
        if aStair.mat_linkstep is not None and aStair.mat_linkstep.IsAlive():
            ttag = step.MakeTag(c4d.Ttexture)
            ttag[c4d.TEXTURETAG_MATERIAL] = aStair.mat_linkstep
            ttag[c4d.TEXTURETAG_PROJECTION] = 6
        
        zpos = sdims.z / 2.0 - aStair.step_dims.z

        step.SetAbsPos(c4d.Vector(xoff, sdims.y / 2.0, zpos))
        max_dims.y += sdims.y
    
    if not aChild_stair.reverse and not only_steps and not aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SOLID:
        wedge_width = aChild_stair.width
        xpos = - platform_dims.x / 2
        
        if aChild_stair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_SINGLE:
            wedge_width = aChild_stair.base_width
            xpos = aChild_stair.width / 2.0 - platform_dims.x / 2.0 - wedge_width / 2.0
        
        wedgeNull = c4d.BaseObject(c4d.Onull)
        wedgeNull.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_PLATFORMOBJ))
        wedgeNull.InsertUnder(platformNull)
        extrude = getExtrude(wedge_width, wedgeNull, aStair.mat_linkbase)
        getSpline(splinegen.generateWedge(wedge_offset, wedge_end, wegde_height, aChild_stair.stair_zigzag), extrude)
        wedgeNull.SetAbsPos(c4d.Vector(xpos, aStair.treads_height, platform_dims.z))
        
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_LEFT:
            wedgeNull.SetAbsPos(c4d.Vector(- platform_dims.x / 2, aStair.treads_height, - wedge_width / 2 + aChild_stair.width / 2 + aStair.platform_depth))
            wedgeNull.SetAbsRot(c4d.Vector(math.pi / 2.0, 0.0, 0.0))
            
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_RIGHT:
            wedgeNull.SetAbsPos(c4d.Vector(platform_dims.x / 2, aStair.treads_height, wedge_width / 2 + aChild_stair.width / 2 + aStair.platform_depth))
            wedgeNull.SetAbsRot(c4d.Vector(- math.pi / 2.0, 0.0, 0.0))
            
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_ULEFT:
            wedgeNull.SetAbsPos(c4d.Vector(xpos + wedge_width, aStair.treads_height, - aStair.step_dims.z))
            wedgeNull.SetAbsRot(c4d.Vector(math.pi, 0.0, 0.0))
            
        if aStair.platform_direction == ids.SMARTSTAIRS_PLATFORM_DIR_URIGHT:
            wedgeNull.SetAbsPos(c4d.Vector(platform_dims.x / 2 - aChild_stair.width / 2 + wedge_width / 2, aStair.treads_height, - aStair.step_dims.z))
            wedgeNull.SetAbsRot(c4d.Vector(math.pi, 0.0, 0.0))
                        
    posx = aStair.width / 2.0 - platform_dims.x / 2.0
    posy = aStair.height - aStair.treads_height
    posz = aStair.depth / 2.0
    if aStair.reverse:
        if aStair.platform_direction != ids.SMARTSTAIRS_PLATFORM_DIR_FRONT:
            posx += aStair.platform_width
        posy = platform_dims.y
        posz = - aStair.depth / 2.0
        platformNull.SetAbsRot(c4d.Vector(- math.pi, 0.0, 0.0))
    if aStair.reverse:
        posy -= solidoff    
    
    platformNull.SetAbsPos(c4d.Vector(posx, posy, posz))
    
    if aStair.mat_linkplatform is not None and aStair.mat_linkplatform.IsAlive():
        ttag = platformNull.MakeTag(c4d.Ttexture)
        ttag[c4d.TEXTURETAG_MATERIAL] = aStair.mat_linkplatform
        ttag[c4d.TEXTURETAG_PROJECTION] = 6

    return platformNull, max_dims


def getNull():

    null = c4d.BaseObject(c4d.Onull)

    null.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_NULLOBJ))

    return null


def getExtrude(aHeight, aParent, aMat, aFlip = False):

    extrude = c4d.BaseObject(c4d.Oextrude)

    extrude.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_EXTRUDEOBJ))
    
    extrude[c4d.EXTRUDEOBJECT_DIRECTION] = 5
    
    if not aFlip:
        extrude[c4d.EXTRUDEOBJECT_MOVE] = c4d.Vector(aHeight, 0.0, 0.0)
    else:
        extrude[c4d.EXTRUDEOBJECT_MOVE] = c4d.Vector(0.0, aHeight, 0.0)

    extrude.InsertUnder(aParent)
    
    if aMat is not None and aMat.IsAlive():
        ttag = extrude.MakeTag(c4d.Ttexture)
        ttag[c4d.TEXTURETAG_MATERIAL] = aMat
        ttag[c4d.TEXTURETAG_PROJECTION] = 6

    return extrude


def getSpline(aSpline, aParent):

    aSpline.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_SPLINEOBJ))
    aSpline.InsertUnder(aParent)

    return aSpline


def getReference(aParent):

    nullReference = c4d.BaseObject(c4d.Onull)

    nullReference.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPREFERENCEOBJ))
    nullReference.SetEditorMode(c4d.MODE_OFF)
    nullReference.SetRenderMode(c4d.MODE_OFF)

    nullReference.InsertUnder(aParent)

    return nullReference


def getStepInstances(aStair, aStepInstance, aParent):

    nullInstances = c4d.BaseObject(c4d.Onull)
    nullInstances.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPINSTANCESOBJ))
    nullInstances.InsertUnder(aParent)
    
    skip_start = aStair.is_parent and aStair.reverse# and aStair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3
    skip_end = (aStair.is_parent and not aStair.reverse) or (aStair.is_child and aStair.reverse)  
    
    yoff = 0
    if (aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or
        aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3):
        yoff = aStair.treads_offset_y        

    for i in range(1 if skip_start else 0, aStair.num_steps - (1 if  skip_end else 0)):
        obj = aStepInstance.GetClone()
        pos = c4d.Vector(0.0, 
              (i + 1) * aStair.step_dims.y - aStair.treads_height / 2.0 - yoff, 
              i * aStair.step_dims.z - aStair.depth / 2.0 + aStair.step_dims.z / 2.0 - aStair.treads_offset_z - aStair.treads_depth / 2.0)
        obj.SetAbsPos(pos)    
        obj.InsertUnder(nullInstances)

    return nullInstances


def getStepInstance(aStair, aParent):

    if aStair.treads_link is None or not aStair.treads_link.IsAlive():
        nullReference = getReference(aParent)
        
        width_off = 0
        if aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3:
            width_off = aStair.base_width * 2
    
        step = c4d.BaseObject(c4d.Ocube)
        
        step[c4d.PRIM_CUBE_LEN] = c4d.Vector(aStair.step_dims.x - width_off, aStair.treads_height, aStair.step_dims.z + aStair.treads_depth)
        step[c4d.PRIM_CUBE_DOFILLET] = aStair.treads_have_fillet
        step[c4d.PRIM_CUBE_FRAD] = aStair.treads_fillet_rad
        step[c4d.PRIM_CUBE_SUBF] = aStair.treads_fillet_sub
    
        step.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPOBJ))
        step.InsertUnder(nullReference)
    else:
        step = aStair.treads_link
        
    if aStair.mat_linkstep is not None and aStair.mat_linkstep.IsAlive():
        ttag = step.MakeTag(c4d.Ttexture)
        ttag[c4d.TEXTURETAG_MATERIAL] = aStair.mat_linkstep
        ttag[c4d.TEXTURETAG_PROJECTION] = 6

    instance = c4d.BaseObject(c4d.Oinstance)
    instance[c4d.INSTANCEOBJECT_LINK] = step
    instance[c4d.INSTANCEOBJECT_RENDERINSTANCE_MODE] = True
    instance.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_STEPINSTANCEOBJ))

    if aStair.treads_link is not None and aStair.treads_link_scale:
        link_dims = aStair.treads_link.GetRad()

        if link_dims != aStair.step_dims and link_dims.x > 0 and link_dims.y > 0 and link_dims.z > 0:
            scale_x = float(aStair.step_dims.x) / float(link_dims.x) / 2.0
            scale_y = float(aStair.treads_height) / float(link_dims.y) / 2.0
            scale_z = float(aStair.step_dims.z + aStair.treads_depth) / float(link_dims.z) / 2.0
            instance.SetAbsScale(c4d.Vector(scale_x, scale_y, scale_z))
    
    return instance


def getHandleInstances(aStair, aStepInstance, aParent):
    
    numsteps = aStair.num_steps
    height = aStair.railing_height
    radius = aStair.railing_rad1
    radius2 = aStair.railing_rad2
    xoff = aStair.railing_offsetx
    zoff = aStair.railing_offsetz
    nth = aStair.railing_nth
    foff = 0
    
    if aStair.reverse and aStair.is_child and aStair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
        numsteps -= 1
    
    if aStair.treads_have_fillet:
        foff = aStair.treads_fillet_rad

    nullInstances = c4d.BaseObject(c4d.Onull)
    nullInstances.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_RAILINGINSTANCESOBJ))
    nullInstances.InsertUnder(aParent)
        
    yoff = 0
    if (aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE2 or
        aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_DOUBLE3):
        yoff = aStair.treads_offset_y        

    pos_s = c4d.Vector()
    pos_e = c4d.Vector()
        
    rad = aStair.width / 2.0 - foff - xoff
    if rad == 0: rad = 0.001
    rad2 = aStair.stair_helix_innerrad + radius + foff + xoff
    if rad2 == 0: rad2 = 0.001
    
    rd = (aStair.stair_helix_deg)/float(numsteps)
    ro = ((zoff + radius + foff - aStair.treads_depth / 2.0) / rad)
    if aStair.stair_helix_deg < 0:
        ro *= -1
    
    for i in range(0, numsteps):
        
        if not i == 0 and not i == numsteps - 1:
            if nth > 1 and (i + 1) % nth != 1:
                continue
                    
        pos = c4d.Vector(- aStair.step_dims.x / 2.0 + foff + xoff + radius, 
                         (i + 1) * aStair.step_dims.y - yoff + height / 2.0, 
                         i * aStair.step_dims.z - aStair.depth / 2.0 - aStair.treads_offset_z - aStair.treads_depth + radius + foff + zoff)
        
        if pos.x > 0: pos.x = 0
        
        if aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_RIGHT:
            pos.x = aStair.step_dims.x / 2.0 - foff - xoff - radius
            if pos.x < 0: pos.x = 0
                
        if aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX: 
            pos.x = - math.sin(rd * i + ro) * rad
            pos.z = math.cos(rd * i + ro) * rad
            if aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_LEFT:
                pos.x = - math.sin(rd * i + ro) * rad2
                pos.z = math.cos(rd * i + ro) * rad2
               
        cyl = aStepInstance.GetClone()
        cyl.SetAbsPos(c4d.Vector(pos))
        cyl.InsertUnder(nullInstances)

        if aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_BOTH:
            cyl2 = cyl.GetClone()
            pos.x = aStair.step_dims.x / 2.0 - foff - xoff - radius
            if pos.x < 0: pos.x = 0
            
            if aStair.stairtype == ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
                pos.x = - math.sin(rd * i + ro) * rad2
                pos.z = math.cos(rd * i + ro) * rad2
                            
            cyl2.SetAbsPos(c4d.Vector(pos))
            cyl2.InsertUnder(nullInstances)
        
        if i == 0:
            pos_s = pos
            
        if i == numsteps - 1:
            pos_e = pos     
                   
    swn = c4d.BaseObject(c4d.Osweep)
    tag = swn.MakeTag(c4d.Tphong)
    tag[c4d.PHONGTAG_PHONG_ANGLELIMIT] = True
    
    if aStair.mat_handrail is not None and aStair.mat_handrail.IsAlive():
        ttag = swn.MakeTag(c4d.Ttexture)
        ttag[c4d.TEXTURETAG_MATERIAL] = aStair.mat_handrail
        ttag[c4d.TEXTURETAG_PROJECTION] = 6
    
    swn.InsertUnder(aParent)
    swn.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_RAILINGSWEEPOBJ))
    
    if aStair.railing_hlink is None or not aStair.railing_hlink.IsAlive():
        cir = c4d.BaseObject(c4d.Osplinecircle)
        cir[c4d.PRIM_CIRCLE_RADIUS] = radius2
    else:
        cir = aStair.railing_hlink.GetClone(c4d.CL_NO_BITS)
        
        if cir is not None and aStair.railing_hlink_scale:
            link_dims = cir.GetRad()
    
            if (link_dims.x != radius2 * 2.0 or link_dims.y != radius2 * 2.0) and link_dims.x > 0 and link_dims.y > 0:
                scale_x = float(radius2) / float(link_dims.x)
                scale_y = float(radius2) / float(link_dims.y)
                cir.SetAbsScale(c4d.Vector(scale_x, scale_y, link_dims.z))
    
    if aStair.mat_linkbase is not None and aStair.mat_linkbase.IsAlive():
        ttag = cir.MakeTag(c4d.Ttexture)
        ttag[c4d.TEXTURETAG_MATERIAL] = aStair.mat_linkbase
        ttag[c4d.TEXTURETAG_PROJECTION] = 6
    
    if aStair.stairtype != ids.SMARTSTAIRS_STAIRS_TYPE_HELIX:
        pos_swn = c4d.Vector(pos_s.x, pos_s.y + height / 2.0, pos_s.z)
    
        swn.SetAbsPos(pos_swn)
        
        spl = c4d.SplineObject(4, c4d.SPLINETYPE_LINEAR)
    
        y1 = aStair.height / aStair.depth * aStair.railing_overbottom
        y2 = aStair.height / aStair.depth * aStair.railing_overtop
        if y1 > aStair.railing_overbottom * 4: y1 = aStair.railing_overbottom * 4
        if y2 > aStair.railing_overtop * 4: y2 = aStair.railing_overtop * 4
        spl.SetPoint(0, c4d.Vector(0, - y1, - aStair.railing_overbottom))
        spl.SetPoint(1, c4d.Vector(0, 0, 0))
        spl.SetPoint(2, c4d.Vector(0, pos_e.y - pos_s.y, pos_e.z - pos_s.z))
        spl.SetPoint(3, c4d.Vector(0, pos_e.y - pos_s.y + y2, pos_e.z - pos_s.z + aStair.railing_overtop))
        spl.InsertUnder(swn)
        
        cir.InsertUnder(swn)
        
        if aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_BOTH:
            swn2 = swn.GetClone()
            pos_swn.x = - aStair.step_dims.x / 2.0 + foff + xoff + radius
            if pos_swn.x > 0: pos_swn.x = 0
            swn2.SetAbsPos(c4d.Vector(pos_swn))
            swn2.InsertUnder(aParent)
    else:                    
        if aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_BOTH or aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_RIGHT:
            helix = c4d.BaseObject(c4d.Osplinehelix)
            helix[c4d.PRIM_HELIX_RADIUS1] = rad
            helix[c4d.PRIM_HELIX_RADIUS2] = rad
            helix[c4d.PRIM_HELIX_HEIGHT] = aStair.height - aStair.step_dims.y
            helix[c4d.PRIM_PLANE] = 2
            helix[c4d.PRIM_HELIX_END] = aStair.stair_helix_deg - (aStair.stair_helix_deg/numsteps)
            helix.SetAbsPos(c4d.Vector(0.0, pos_s.y + height / 2.0, 0.0))
            helix.SetAbsRot(c4d.Vector(math.pi / 2.0 + ro, 0.0, 0.0))
            helix.InsertUnder(swn)
            
            cir.InsertUnder(swn)
        
        if aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_BOTH or aStair.railing_type == ids.SMARTSTAIRS_RAILING_TYPE_LEFT:
            swn2 = c4d.BaseObject(c4d.Osweep)
            tag = swn2.MakeTag(c4d.Tphong)
            tag[c4d.PHONGTAG_PHONG_ANGLELIMIT] = True
            
            swn2.InsertUnder(aParent)
            swn2.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_RAILINGSWEEPOBJ))
            
            helix = c4d.BaseObject(c4d.Osplinehelix)
            helix[c4d.PRIM_HELIX_RADIUS1] = rad2
            helix[c4d.PRIM_HELIX_RADIUS2] = rad2
            helix[c4d.PRIM_HELIX_HEIGHT] = aStair.height - aStair.step_dims.y
            helix[c4d.PRIM_PLANE] = 2
            helix[c4d.PRIM_HELIX_END] = aStair.stair_helix_deg - (aStair.stair_helix_deg/numsteps)
            helix.SetAbsPos(c4d.Vector(0.0, pos_s.y + height / 2.0, 0.0))
            helix.SetAbsRot(c4d.Vector(math.pi / 2.0 + ro, 0.0, 0.0))
            helix.InsertUnder(swn2)
            
            cir = cir.GetClone(c4d.CL_NO_BITS)
            cir.InsertUnder(swn2)

    return nullInstances


def getHandleInstance(aStair, aParent):

    height = aStair.railing_height
    radius = aStair.railing_rad1
        
    if aStair.railing_link is None or not aStair.railing_link.IsAlive():
        nullReference = getReference(aParent)
        nullReference.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_RAILINGREFERENCEOBJ))
        
        cyl = c4d.BaseObject(c4d.Ocylinder)
        tag = cyl.MakeTag(c4d.Tphong)
        tag[c4d.PHONGTAG_PHONG_ANGLELIMIT] = True
        
        cyl[c4d.PRIM_CYLINDER_RADIUS] = radius
        cyl[c4d.PRIM_CYLINDER_HEIGHT] = height
        cyl.InsertUnder(nullReference)
       
        cyl.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_RAILINGOBJ))
    else:
        cyl = aStair.railing_link
        
    if aStair.mat_linkrailing is not None and aStair.mat_linkrailing.IsAlive():
        ttag = cyl.MakeTag(c4d.Ttexture)
        ttag[c4d.TEXTURETAG_MATERIAL] = aStair.mat_linkrailing
        ttag[c4d.TEXTURETAG_PROJECTION] = 6

    instance = c4d.BaseObject(c4d.Oinstance)
    instance[c4d.INSTANCEOBJECT_LINK] = cyl
    instance[c4d.INSTANCEOBJECT_RENDERINSTANCE_MODE] = True
    instance.SetName(plugins.GeLoadString(ids.STR_SMARTSTAIRS_RAILINGINSTANCEOBJ))

    if aStair.railing_link is not None and aStair.railing_link_scale:
        link_dims = aStair.railing_link.GetRad()

        if link_dims != c4d.Vector(aStair.railing_rad1 * 2) and link_dims.x > 0 and link_dims.y > 0 and link_dims.z > 0:
            scale_x = float(aStair.railing_rad1) / float(link_dims.x)
            scale_y = float(aStair.railing_height / 2.0) / float(link_dims.y)
            scale_z = float(aStair.railing_rad1) / float(link_dims.z)
            instance.SetAbsScale(c4d.Vector(scale_x, scale_y, scale_z))
    
    return instance